#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "biquad.h"

typedef struct cx_t {
    float r;  //magnitude or //real
    float i;  //phase or //imaginary
} cx;

biquad_coeffs* 
make_biquad(int type, biquad_coeffs* cf, float fs, float f0, float Q)
{
    float w0 = 2.0*M_PI*f0/fs;
    float c = cos(w0);
    float s = sin(w0);
    float alpha = s/(2.*Q);
    float a0, a1, a2;
    float b0, b1, b2;
    float g;
    cf = (biquad_coeffs*) malloc(sizeof(biquad_coeffs));
    

    switch(type){
        case LPF:
            b0 =  (1.0 - c)/2.0;
            b1 =   1.0- c;
            b2 =  (1.0 - c)/2.0;
            a0 =   1.0 + alpha;
            a1 =  -2.0*c;
            a2 =   1.0 - alpha;
            break;
            
         case HPF:
            b0 =  (1.0 + c)/2.0;
            b1 =   -(1.0 + c);
            b2 =  (1.0 + c)/2.0;
            a0 =   1.0 + alpha;
            a1 =  -2.0*c;
            a2 =   1.0 - alpha;
            break;
            
         case BPF:
            b0 =   Q*alpha;
            b1 =   0.0;
            b2 =  -Q*alpha;
            a0 =   1.0 + alpha;
            a1 =  -2.0*c;
            a2 =   1.0 - alpha;
            break;
         case LPF1P:
            //1-pole high pass filter coefficients
            // H(z) = g * (1 - z^-1)/(1 - a1*z^-1)
            // Direct Form 1: 
            //    h[n] = g * ( b0*x[n] - b1*x[n-1] ) - a1*y[n-1]
            // In below implementation gain is redistributed to the numerator:
            //    h[n] = gb0*x[n] - gb1*x[n-1] - a1*y[n-1]
            a0 = 1.0;
            a2 = 0.0;
            a1 = -expf(-w0);
            g = (1.0 + a1)/1.12; //0.12 zero improves RC filter emulation at higher freqs.
            b0 = g;
            b1 = 0.12*g;
            b2 = 0.0;
            break;
         case HPF1P:
            //1-pole high pass filter coefficients
            // H(z) = g * (1 - z^-1)/(1 - a1*z^-1)
            // Direct Form 1: 
            //    h[n] = g * ( x[n] - x[n-1] ) - a1*y[n-1]
            // In below implementation gain is redistributed to the numerator:
            //    h[n] = g*x[n] - g*x[n-1] - a1*y[n-1]
            a0 = 1.0;
            a2 = 0.0;
            a1 = -expf(-w0);
            g = (1.0 - a1)*0.5;
            b0 = g;
            b1 = -g;
            b2 = 0.0;
            break;
                     
        default:
            break;          
    }

    b0 /=  a0;
    b1 /=  a0;
    b2 /=  a0;  
    a1 /=  a0;
    a2 /=  a0;
    
    cf->b0 = b0;
    cf->b1 = b1;
    cf->b2 = b2;
    cf->a1 = -a1;  // filter implementation uses addition instead of subtraction
    cf->a2 = -a2;  // so "a" coefficients must be negated to compensate
          
    cf->y1 = 0.0;
    cf->y2 = 0.0;
    cf->x1 = 0.0;
    cf->x2 = 0.0;
    
    return cf;

}

float*
make_butterworth_coeffs(int order, float* coeffs)
{
    coeffs = (float*) malloc(sizeof(float)*order);
    
    int k = 1;
    int n = order;
    float fn = (float) n;
    
    //printf("N, Coeff\n", k, coeffs[k-1]);
    
    if(n%2 == 0) {
        for(k=1; k<=n/2; k++)
        {
            float fk = (float) k;
            coeffs[k-1] = 1.0/(-2.0*cos( (2.0*fk + fn -1)/(2.0*fn) * M_PI)); //1/x returns filter stage Q factor
            //printf("%d, %f\n", k, coeffs[k-1]);
        }
    } else { //odd
        for(k=1; k<=(n-1)/2; k++)
        {
            float fk = (float) k;
            coeffs[k-1] = 1.0/( -2.0*cos( (2.0*fk + fn -1)/(2.0*fn) * M_PI) ); //1/x returns filter stage Q factor
            //printf("%d, %f\n", k, coeffs[k-1]);
        } 
    }
    
    return coeffs;
}

float
run_filter(float x, biquad_coeffs* cf)
{

    float y0 = cf->b0*x + cf->b1*cf->x1 + cf->b2*cf->x2
                        + cf->a1*cf->y1 + cf->a2*cf->y2;
    cf->x2 =cf->x1;
    cf->x1 = x;
    cf->y2 = cf->y1;
    cf->y1 = y0;
    return y0;
                      
}

float
run_filter_one_pole(float x, biquad_coeffs* cf)
{

    float y0 = cf->b0*x + cf->b1*cf->x1
                        + cf->a1*cf->y1;
    cf->x1 = x;
    cf->y1 = y0;
    return y0;
                      
}

//1-pole all-pass filter takes HPF as input.
// Nonlinear distortion such as found in a FET phaser can be applied
// by modulating the a1 coefficient with signal level.
float
run_APF_one_pole(float x, biquad_coeffs* cf)
{

    float y0 = cf->b0*x + cf->b1*cf->x1
                        + cf->a1*cf->y1;
    cf->x1 = x;
    cf->y1 = y0;
    return 2.0*y0 - x; //converts to APF function
                      
}

float sqr(float x)
{
    return x*x;
}

void plot_response(float f1, float f2, int pts, biquad_coeffs* cf, float fs, cx *r)
{
    float dp = 2.0*M_PI;
    // z^-1 = cos(w) - j*sin(w)
    // z^-2 = cos(2*w) - j*sin(2w)
    
    float num[2], den[2];
    num[0] = num[1] = 0.0;
    den[0] = den[1] = 0.0;
    
    float dw = dp*(f2-f1)/pts;
    dw /= fs;
    float w = dp*f1/fs;
    float magn = 0.0;
    float magd = 0.0;
    float an = 0.0;
    float ad = 0.0;
    
    int n = abs(pts) + 1;
    
    int i;
    for(i=0; i<n; i++)
    {
        num[0] = cf->b0 + cf->b1*cos(-w) + cf->b2*cos(-2.0*w); //numerator real part
        num[1] = cf->b1*sin(-w) + cf->b2*sin(-2.0*w);
        den[0] = 1.0 - cf->a1*cos(-w) - cf->a2*cos(-2.0*w); //numerator real part
        den[1] = -cf->a1*sin(-w) - cf->a2*sin(-2.0*w); //negated because of filter implementation
        
        magn = sqrtf(sqr(num[0]) + sqr(num[1]));
        magd = sqrtf(sqr(den[0]) + sqr(den[1]));
        an = atanf(num[1]/num[0]); //angle of numerator
        ad = atanf(den[1]/den[0]); //angle of denominator
        
        r[i].r = magn/magd;
        r[i].i = an - ad;
        
        //printf("%f\t%f\t%f\n", fs*w/dp, 20.0*log10(r[i].r), 180.0*r[i].i/(M_PI));
        
        w += dw;
    }


}

int main(int argc, char** argv)
{
    float *q;
    // int tb = 0;
    // for(tb=0; tb<8; tb++) {
    // printf("\nORDER: %d\n", tb+1);
    q = make_butterworth_coeffs(8, q);
    // free(coeffs);
    // }
   // return 0;
    
    float fs = 44100.0;
    float fcut = 7200.0;
    float fcut2 = 7200.0;
    
    //complex representation of filter response
    //cx types are treated as phasors (A*e^j_theta), not rectangular (r + ji)
    int N = 1000;
    float fl = 20.0;  //output plot start frequency
    float fh = 20000.0;  //output plot end frequency
    cx* response1 = (cx*) malloc(sizeof(cx)*(N+1));
    cx* response2 = (cx*) malloc(sizeof(cx)*(N+1));
    cx* response3 = (cx*) malloc(sizeof(cx)*(N+1));
    cx* response4 = (cx*) malloc(sizeof(cx)*(N+1));
    cx* response5 = (cx*) malloc(sizeof(cx)*(N+1));
    
    //construct filters
    biquad_coeffs* f[5];
    f[0] = make_biquad(LPF, f[0], fs, fcut, q[0]);
    f[1] = make_biquad(LPF, f[1], fs, fcut2, q[1]);
    f[2] = make_biquad(LPF, f[2], fs, fcut, q[2]);   
    f[3] = make_biquad(LPF, f[3], fs, fcut2, q[3]);
    f[4] = make_biquad(HPF1P, f[4], fs, 750.0, 1.0); 

    //analyze filters
    plot_response(fl, fh, (float) N, f[0], fs, response1);
    plot_response(fl, fh, (float) N, f[1], fs, response2);
    plot_response(fl, fh, (float) N, f[2], fs, response3);
    plot_response(fl, fh, (float) N, f[3], fs, response4);
    plot_response(fl, fh, (float) N, f[4], fs, response5);

    //output composite filter response
    int j;
    float mag = 0.0;
    float phase = 0.0;
    float fn = (float) N;
    float df = (fh - fl)/fn;
    for (j=0; j<(N+1); j++)
    {
    /*
        mag = response1[j].r * response2[j].r * response3[j].r * response4[j].r;
        phase = response1[j].i + response2[j].i + response3[j].i + response4[j].i;
    */    
        //need to rename .r and .i because plot_response returns mag and phase
        mag = response5[j].r;
        phase = response5[j].i;
        //printf("%f\t%f\t%f\n", fl + df*((float) j), 20.0*log10(mag), 180.0*phase/(M_PI)); 
    }
    
    float y0 = 0.0;
    float x0 = 0.0;
    int maxtime = ((int)2.0*fs/750.0);
    int startstep = maxtime/10;
    printf("Time\ty0\n");
    for(j=0; j<maxtime; j++)
    {
        if(j>startstep)
            x0 = 1.0;
        y0 = run_APF_one_pole(x0, f[4]);
        printf("%f\t%f\n", ((float) j)/fs, y0);
    }
    
    for(j=0;j<5;j++)
    {
        free(f[j]);
    }
    free(q);
    
    return 0;
}




